import log
import os
import time
#
# Generic OpalKelly based controller interface
#
# See http://www.opalkelly.com/library/FrontPanelAPI/
from opal_kelly import *

def is_pow2(n):
	return n >= 1 and not (n & (n - 1))

def round_to_pow2(n):
	if is_pow2(n):
		return n
	sht = 1
	while True:
		n_ = n >> sht
		if not n_:
			break
		n |= n_
		sht *= 2
	n2 = n + 1
	assert is_pow2(n2)
	return n2

class ControllerError(Exception):
	def __init__(self, what):
		Exception.__init__(self, what)

class ok_ctl:
	"""Opal Kelly firmware loader"""
	def __init__(self, cfg):
		"""The cfg parameter receive the implementation specific parameters namespace"""
		self.cfg = cfg

	def init_core(self, port = ''):
		"""Init core library"""
		self.dev = FrontPanel()
		if self.dev.NoError != self.dev.OpenBySerial(port):
			return False
		return True

	def init_fw(self, force_load):
		"""Load firmware if necessary"""
		if force_load or not self.dev.IsFrontPanelEnabled():
			log.notice('loading controller firmware')
			if self.dev.NoError != self.dev.LoadDefaultPLLConfiguration():
				raise ControllerError('failed to init PLL')
			path = os.path.join(os.path.dirname(__file__), self.cfg.fw_path)
			if not os.path.exists(path):
				raise ControllerError('failed to locate firmware %s' % path)
			log.dbg('using firmware %s' % path)
			if self.dev.NoError != self.dev.ConfigureFPGA(path):
				raise ControllerError('failed to init FPGA')
		else:
			log.dbg('firmware is already loaded')
		if not self.dev.IsFrontPanelEnabled():
			raise ControllerError('front panel is not supported by the firmware')

	def init_controller(self, force_load = False, port = ''):
		"""Open Opel Kelly device and initialize it if necessary"""
		if not self.init_core(port):
			return False
		if hasattr(self.cfg, 'on_load'):
			self.cfg.on_load(self)
		self.init_fw(force_load)
		return True

	def get_id(self):
		"""Returns id string of the controller hardware"""
		return self.dev.GetDeviceID()

	def read_buff(self, ep, buff):
		"""Read data from pipe to the buffer"""
		cnt = self.dev.ReadFromPipeOut(ep, buff)
		if cnt != len(buff):
			raise ControllerError('failed to read %d bytes: %d returned' % (len(buff), cnt))

	@staticmethod
	def nz_word_index(buff):
		"""Returns first non-zero word offset"""
		off = (i for i, c in enumerate(buff) if c).next()
		if off & 1:
			off -= 1
		return off

##	@staticmethod
	def find_first_word(self, buff, start_byte = None):
		"""Returns first non-zero word offset given the optional start byte hint"""
		if start_byte is None:
			if not any(buff):
				return -1
			return self.nz_word_index(buff)
		else:
			try:
				return buff.index(chr(start_byte))
			except ValueError:
				pass
			if any(buff):
				return self.nz_word_index(buff)
			return -1

	def read_chunk_skipz(self, ep, sz, msec_tout = None, start_byte = None):
		"""Read data string from the given pipe endpoint skipping leading zeros"""
		assert sz > 0 and sz % 2 == 0
		if msec_tout:
			deadline = time.time() + msec_tout/1000.0
		res, buff = bytearray(), bytearray(sz)
		while True:
			self.read_buff(ep, buff)
			if res:
				res += buff
				break
			skip = self.find_first_word(buff, start_byte)
			if skip >= 0:
				res = buff[skip:]
				if not skip:
					break
				buff = buff[:skip]
			elif msec_tout and time.time() > deadline:
				raise ControllerError('timeout (%d msec) waiting data (%d bytes)' % (msec_tout, len(buff)))

		assert len(res) == sz
		return res

	def read_chunk_stream(self, ep, sz, grp, cb, msec_tout = None, start_byte = None):
		"""Read continuous stream of data chunks calling callback"""
		assert sz > 0 and sz % 2 == 0
		assert grp >= 1
		if msec_tout:
			deadline = time.time() + msec_tout/1000.0
		buff, tail = None, None
		buff_sz = round_to_pow2(sz * grp)
		while True:
			if buff is None:
				buff = bytearray(buff_sz)
				self.read_buff(ep, buff)
			if tail is not None:
				# We have the beginning of the chunk left from the previous buffer
				buff = tail + buff
				tail = None
				skip = 0
			else:
				# Lookup data chunk
				skip = self.find_first_word(buff, start_byte)
			if skip >= 0: # Chunk found
				if msec_tout:
					deadline = time.time() + msec_tout/1000.0
				if skip + sz > len(buff):
					# The chunk is bigger than remaining buffer space
					tail = buff[skip:]
					buff = None
				else:
					cb(buff[skip:skip+sz])
					if skip+sz < len(buff):
						# Truncate buffer
						buff = buff[skip+sz:]
					else:
						# No more data in the buffer
						buff = None
			else: # No chunks in buffer
				if msec_tout and time.time() > deadline:
					raise ControllerError('timeout (%d msec) waiting data chunk' % (msec_tout))
				buff = None
