#!/usr/bin/env python3
from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from builtins import map
from builtins import str
from builtins import range
from builtins import object
from past.utils import old_div
from importlib import import_module
import threading 
import numpy as np
import cv2
import zipfile, os, sys, tempfile
import io

from .iface import *

from . import ref as ref

import sys

from time import sleep

NOPROCESS   =   0x01
STITCH     =   0x08
GREEDYSTITCH  =   0x40
ACCUMULATE  =   0x80
DIFFERENCE  =   0x10
DEFAULT_FLAGS = STITCH | GREEDYSTITCH

DEFAULT_BG_FILENAMES = ["bg/0.bg","bg/1.bg","bg/2.bg","bg/3.bg","bg/4.bg","bg/5.bg","bg/6.bg","bg/7.bg","bg/8.bg","bg/9.bg","bg/10.bg"]
DEFAULT_NORM_KEYS = ("default","recorded")
DEFAULT_NORM_FILENAMES = {"default":"norm/default.nm","recorded":"norm/recorded.nm"}

RAW_MAX = 0.7


class AutoModeException(Exception):
    def __init__(self, stop = True):
        self.stop = stop
    

        
class progress(object):
    def __init__(self,total=100,callback=None):
        """Helper class for managing progress of a process.
    total - length of the process in arbitrary units (default 100)
    callback - optional function to call back on change with percentage of the total as an argument."
"""
        self.current = 0.0
        self.total = float(total)
        self.progress = old_div(self.current,self.total)
        if callable(callback):
            self.callback = callback
            self.callback(0)
        else:
            self.callback = None
    def increase(self,value=1):
        """Augment progress by the value.
    value - size of the increment (default 1)
"""
        self.current += value
        if self.current > self.total:
            self.current = self.total
        self.progress = old_div(self.current,self.total)
        if self.callback is not None:
            return self.callback(self.progress*100)
        return True, False
    
    def get_progress(self):
        """Get current progress in percents of total"""
        return self.progress*100

    

class processor(object):
    """Main class for data acquisition and processing."""


  
    def __init__(self,threaded=True, config=None, defaults=None, flags = DEFAULT_FLAGS, sn=''):
        """Creates interface to the hardware and setups acquisition and data processing.
    Parameters:
        threaded    -   whether to use separate threads (via threading modeule) for aquisition and processing or not (boolean, default: True)
        config      -   path to a configuration file containing background and normalization data (default: None)
        defaults    -   path to a defaults file
        flags       -   data processing flags (default: DEFAULT_FLAGS)
        sn          -   serial number string of the device to open (default: empty string=the first available)

    Data processing flags:
        STITCH      -   heal over isolated missing pixels
        GREEDYSTITCH-   heal over large area of missing pixels
        ACCUMULATE  -   turn on accumulation (time-domain filtering)
        DIFFERENCE  -   turn on differential signal between consecutive frames
        DEFAULT_FLAGS = STITCH | GREEDYSTITCH
"""

        if defaults is None:
##            defaults = __import__("terasense.tbus.config",globals(),locals(),['config']).config
            defaults = import_module(".config","terasense.config")
        if config is None:
            try:
                config = os.path.join(os.path.dirname(__file__),defaults.configPath)
            except:
                config = ""
       
        self.threaded = threaded
        self.interface = interface(threaded, sn)

        self.X_SIZE = self.interface.X_SIZE
        self.Y_SIZE = self.interface.Y_SIZE

        self.processFlags = flags

        
        self._want_setexp = None

        self._callback = None

        self.currentNorm = None
        self.currentBG = None
        
        self.buffer =  np.empty((self.X_SIZE,self.Y_SIZE),dtype=np.float32)
        self.result =  np.empty((self.X_SIZE,self.Y_SIZE),dtype=np.float32)
        self.temp =  np.empty((self.X_SIZE,self.Y_SIZE),dtype=np.float32)

        self.ROI = None                     #region of interest ( (x0,y0), (x1,y1) ), x0 <=x1, y0 <= y1


        #import config.proc as defaults


        if os.path.isfile(config):
            configZipName = config
        elif os.path.isfile(os.path.join(config,self.interface.get_id()+".zip")):
            configZipName = os.path.join(config,self.interface.get_id()+".zip")
        else:
            configZipName = os.path.join(os.path.dirname(__file__),"data/"+self.interface.get_id()+".zip")
        
        

        if self.threaded:
            self.datalock = threading.Lock()
            self.ready = threading.Event()
            self.raw   = threading.Event()            
            self.next_frame  = threading.Event()
            self.acq_thread = None
            self.proc_thread = None
            self._stop_acquisition = False
            self._stop_processing = False

        self.bg = self.bgNone = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=np.float32)
        self.norm = self.normNone = np.ones((self.X_SIZE,self.Y_SIZE),dtype=np.float32)
        self.mask = self.maskNone = 1e99*np.ones((self.X_SIZE,self.Y_SIZE),dtype=np.float32)
        self.stitchList = []
        self.odd_frame = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=np.float32)


        try:
            self.threshold = defaults.threshold # limiting signal/noise ratio during calibration for mask
        except AttributeError:
            self.threshold = 10

        try:
            self.recordingCount = defaults.recordingCount # limiting signal/noise ratio during calibration for mask
        except AttributeError:
            self.recordingCount = 30


            
        try: 
            self.LoadConfig(configZipName)
        except FileNotFoundError as e:
            print("Failed to load config "+configZipName+" :"+str(e),file=sys.stderr)
                  
            self.bgList = []
            self.normDict = {}
            self.MakeNullBGList()
            self.MakeNullNormList()
            self.SelectNorm('none')

        try:
            self.accuLength = max(defaults.accuLength,1) #1-100
        except AttributeError:
            self.accuLength = 10

        try:
            if (defaults.accumulation):
                self.processFlags ^= ACCUMULATE
        except AttributeError:
            pass

        try:
            if (defaults.difference):
                self.processFlags ^= DIFFERENCE
        except AttributeError:
            pass
        
        self.stored = self.bgNone 
        self._reset = True

        self.parity = True
        
        try:
            self.SetExposure(defaults.exposure)
        except AttributeError:
            self.SetExposure(0)
        
                        
    def acquisition(self,callback=None,errorcallback=None):
        """Target for acquisition thread"""
        def cb():
            

            if self._stop_acquisition:
                raise AutoModeException(True)
            if self._want_setexp is not None:
                raise AutoModeException(False)

            self.raw.set()


            self.next_frame.wait(30.0)
            self.next_frame.clear()


        while not self._stop_acquisition:
            try:
                self.interface.auto2buffer(self.buffer, cb)
               
            except RuntimeError as e:
                if callable(errorcallback):
                    errorcallback(e)
                self._stop_processing = True
            except AutoModeException as e:
                if not e.stop:
                    self.reconfigure()
                    self.interface.reset()
                    self.parity = True                    
                    self.ResetAccumulation()
                        
        self._stop_acquisition = False


            
    def processing(self,callback=None,errorcallback=None):
        """Target for data processing thread"""
        call = callable(callback)
        while not self._stop_processing:

            if self.raw.wait(30.0):
                self.interface.lock.acquire()
                data = self.buffer.copy()
                self.interface.lock.release()
                self.raw.clear()
                self.next_frame.set()
            else:
                self._stop_acquisition = True
                self.next_frame.set()
                break
            
            if self._stop_processing:
                break
            self.parity = not self.parity
            temp = data.copy()
            if (self.processFlags & DIFFERENCE):
                
                
##                np.subtract(data,self.odd_frame,data)
##                np.fabs(data,data)
                if self.parity:
                    np.subtract(data,self.odd_frame,data)
                else:
                    np.subtract(self.odd_frame,data,data)
            self.odd_frame = temp
                    

            if not (self.processFlags & NOPROCESS):
                data = self.process(data)

            self.datalock.acquire()
            self.result = data.copy()
            self.ready.set()
            self.datalock.release()
            
           
            if call and not self._stop_processing:
                self._stop_processing = self._stop_acquisition = callback(self.result)
                
            
            
        self._stop_processing = False
        

    def start(self,callback=None,errorcallback=None,resume=False):
        """Start multi-threaded acquisition and processing.
    Parameters:
        callback    -   a target to be called on each aquisition cycle as callback(data), where data is a numpy.ndarray containing acquired data [0,1] (default: None)
        errorcallback - a target to be called in case of an error as errorcallback(error, critical), where "error" is an instance of Exception, and "critical" is a boolean (default: None)
        resume  -   if True and callback is None the value of callback from previous call for the function will be used (default: False).
"""
        if callable(callback):
            self._callback = callback
        elif resume and callable(self._callback):
            callback = self._callback
        else:
            self._callback = None

##        self.odd_frame = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=np.float32)
        
        if self.threaded:
            if self.acq_thread is None or not self.acq_thread.is_alive():
                self.interface.reset()
                self.acq_thread = threading.Thread(target = self.acquisition,args=(None,errorcallback))
                self.raw.clear()
                self.acq_thread.start()
            if self.proc_thread is None or not self.proc_thread.is_alive():
                self.proc_thread = threading.Thread(target=self.processing,args=(self._callback,errorcallback))
                self.ready.clear()
                self.proc_thread.start()
            self.parity = True

            if self.processFlags & DIFFERENCE:
                self.read().mean()
                self.read().mean()

            

    def stop(self, join=True):
        """Stop multi-threaded acquisition and processing.
    Parameters:
        join    -   if True, it will try to join acquisition and processing threads. 
"""
        if self.threaded:

            self._stop_processing = True
            self._stop_acquisition = True
            self.ready.set()
            self.raw.set()
            if join and self.acq_thread is not None and self.acq_thread.is_alive():                
                self.acq_thread.join(5.0)                
            if join and self.proc_thread is not None and self.proc_thread.is_alive():                
                self.proc_thread.join(5.0)
            if join:
                self.interface.reset()

               

    def isRunning(self):
        return self.acq_thread is not None and self.proc_thread is not None and self.acq_thread.is_alive() and self.proc_thread.is_alive()
    
    def read(self):
        """Reads processed data. Returns numpy.ndarray with acquired data or None if in the threading mode but acquisition is not running."""
        if self.threaded:
            if self.proc_thread is not None and self.proc_thread.is_alive():
                if self.ready.wait(30.0):
                    self.datalock.acquire()
                    data = self.result.copy()
                    self.ready.clear()
                    self.datalock.release()
                else:
                    return None
            else:
                return None
        else:
            if (self.processFlags & ACCUMULATE) :
                count = self.GetAccuLength()
            else:
                count = 1
            if (self.processFlags & DIFFERENCE) :
                odd_frame = self.interface.read()
            for i in range(count):
                data = self.interface.read()
 #               print("Pre-Pre-Processor data:") 
                if (self.processFlags & DIFFERENCE) :
                    tmp = data.copy()
                    np.subtract(odd_frame,data, data)
                    odd_frame = tmp
                    

                data = self.process(data)             

        return data

    def read_raw(self):
        """Reads unprocessed data. Returns numpy.ndarray with acquired data or None if in the threading mode but acquisition is not running."""
        if self.threaded:
            if self.proc_thread is not None and self.proc_thread.is_alive():
                if self.ready.wait(30.0):
                    self.interface.lock.acquire()
                    data = self.buffer.copy()
                    self.ready.clear()
                    self.interface.lock.release()
                else:
                    return None
            else:
                return None
        else:
            data = self.interface.read()      
        return data

    def readBGData(self):
        """ Returns raw data for background recording (no processing). If in the threading mode, acquisition should NOT be running."""
        data = self.interface.read()
        return data

    def readNormData(self):
        """ Returns raw data for normalization recording (background substracted). If in the threading mode, acquisition should NOT be running."""
        data = self.interface.read()
        data = self.SubstractBG(data)
        return data

    # processing functions
    def process(self, data):
        """ Process data. """
#        print("process data:", data)
        if (self.processFlags & ACCUMULATE) :
                data = self.Accumulate(data)

        if not (self.processFlags & DIFFERENCE) :
            data = self.SubstractBG(data)
            
        data = self.Normalize(data)
        if (self.processFlags &  STITCH) :
                data = self.Stitch(data)
        
        return data

    def Accumulate(self, data):
        """ Return accumulated data (also changes input data) """
        r = 1- 1/float(self.accuLength)
        if r < 0:
            r = 0
        elif r > 1:
            r = 1
        n = 1-r
        if self._reset:
            self._reset = False
        else:
            np.multiply(data,n,data)
            np.multiply(self.stored,r,self.stored)
            np.add(data,self.stored,data)
        self.stored=data.copy()
        return data

    def SubstractBG(self, data):
        np.subtract(data,self.bg,data)
        return data

    def Normalize(self, data):
        np.true_divide(data,self.norm,data)
##        if self.norm is self.normNone:
##        np.absolute(data,data)
        np.clip(data,0,1,data)
        return data

    def SetStitch(self, on = True):
        """Turns on stitching (healing over isolated missing pixels) on or off."""
        if on:
            self.processFlags |= STITCH
        else:
            self.processFlags &= ~STITCH

    def SetGreedyStitch(self, on = True):
        """Turns on greedy stitching (healing over large area of pixels) on or off. Has no effect if stitching is off."""
        if on:
            self.processFlags |= GREEDYSTITCH
        else:
            self.processFlags &= ~GREEDYSTITCH
        
    def Stitch(self, data):
        """ Heals data over missing pixels according to flags set by SetStitch, SetGreedyStitch. """

        if len(self.stitchList)==0:
            return data
    
        greedy = self.processFlags & GREEDYSTITCH
        if self.ROI is None:
            minX = 0
            minY = 0
            maxX = self.X_SIZE
            maxY = self.Y_SIZE
        else:
            minX = self.ROI[0][0]
            minY = self.ROI[0][1]
            maxX = self.ROI[1][0]+1
            maxY = self.ROI[1][1]+1
        noValue=-1000
        deferred=[]
        
        np.multiply(data, self.stitchMask, data)
        
        cv2.blur(data,(3,3),self.temp,borderType=cv2.BORDER_CONSTANT)
##        cv2.GaussianBlur(data,(3,3),0,self.temp)        
    
        for i,j,r in self.stitchList:
            data[i,j] = old_div(self.temp[i,j]*9,r)        
                
        if greedy :
            deferred= self.greedyList
            for i,j in deferred:
                data[i,j] = noValue
            while len(deferred) > 0:
                newdef =[]
                res = []
                for i,j in deferred:
                    n = 0
                    dat = 0
                    for k in range(i-1,i+2) :
                        for l in range(j-1,j+2) :
                            if ( k >= minX and k < maxX and l >= minY and l< maxY ):
                                if not (k==i and l==j) and data[k,l] != noValue :
                                    n+=1
                                    dat += data[k,l]
                    if n > 0 :
                        res.append((i,j, old_div(dat, n)))
                    else:
                        newdef.append((i,j))
                for i, j, x in res:
                    data[i,j]=x

                if len(deferred)==len(newdef):
                    for i,j in newdef:
                        data[i,j]=0
                    break
                deferred = newdef


        return data

    def SetROI(self,roi=None):
        """Sets/resets region of interest in the form of ((x0,y0),(x1,y1)). Parameters have to satisfy conditions x0<=x1, y0<=y1. """
        try:
            if roi[0][0]<0:
                roi[0][0] = 0
            elif roi[0][0]>=self.X_SIZE:
                roi[0][0] = self.X_SIZE-1            
            if roi[1][0]<0:
                roi[1][0] = 0
            elif roi[1][0]>=self.X_SIZE:
                roi[1][0] = self.X_SIZE-1
                
            if roi[0][1]<0:
                roi[0][1] = 0
            elif roi[0][1]>=self.Y_SIZE:
                roi[0][1] = self.Y_SIZE-1            
            if roi[1][1]<0:
                roi[1][1] = 0
            elif roi[1][1]>=self.Y_SIZE:
                roi[1][1] = self.Y_SIZE-1    
            
            self.ROI = ( (min(roi[0][0],roi[1][0]),min(roi[0][1],roi[1][1])), (max(roi[0][0],roi[1][0]),max(roi[0][1],roi[1][1])) )
        except Exception as e:
            self.ROI = None        
        self.prepareStitch()

    #helper functions
    
    def prepareNorm(self):
        """Prepare data to be used in normalization."""
        try:
            self.norm = self.normSrc.copy()
        except:
            self.norm = np.array(self.normSrc,dtype=np.float32)

        self.norm = np.where(  self.mask >  self.threshold,self.norm, np.inf )
                    
    def prepareStitch(self):
        """Prepare data to be used in stitching operation."""

        
        self.stitchList = []
        self.greedyList = []

        if self.mask is self.maskNone:
            self.stitchMask = np.ones_like(self.mask)
            return
        
        if self.ROI is None:
            minX = 0
            minY = 0
            maxX = self.X_SIZE
            maxY = self.Y_SIZE
        else:
            minX = self.ROI[0][0]
            minY = self.ROI[0][1]
            maxX = self.ROI[1][0]+1
            maxY = self.ROI[1][1]+1
            
        threshold = np.ones_like(self.mask)* np.inf

        threshold[minX:maxX,minY:maxY] = self.threshold
        
        self.stitchMask = np.where(  self.mask >  threshold, 1, 0 )
        
        
        tmp = np.where( self.stitchMask[minX:maxX,minY:maxY] == 0)

        
        for i in range(len(tmp[0])):
            k = tmp[0][i]+minX
            l = tmp[1][i]+minY            
            n = 0
            m = 0
            for x in range(k-1,k+2):
                for y in range(l-1,l+2):
                    if ( x >= minX and x < maxX and y >= minY and y< maxY):
                        n+=1
                        m+=self.stitchMask[x,y]
						
            if (m>0):
                self.stitchList.append((k,l,float(m)))
            else:
                self.greedyList.append((k,l))
            
    def SetExposure(self,exp):
        """Changes current exposure in a thread-aware way.
    Parameters:
        exp -   new exposure value.
    In case of non-threading operation of when thread is not running returns None for success or Exception instance for failure;
    in case of running thread always returns None and in case of failure errorcallback is called by processing thread.
"""
        if self.threaded and self.proc_thread is not None and self.proc_thread.is_alive():
            self._want_setexp = exp
            return None
        else:
            return self._setexp(exp)


    def _setexp(self,exp):
        """Low-level (not thread-aware) change in current exposure.
    Parameters:
        exp -   new exposure value.
    Returns:
        None for success or Exception instance for failure.
"""
        error = self.interface.SetExposure(exp)
        self.SelectBG(exp)            
        self._want_setexp = None
        return error
            

    def GetExposure(self):
        """Gets current exposure."""
        return self.interface.GetExposure()

    def GetExposureRange(self):
        """Gets available exposure range as a tuple (min, max)."""
        return self.interface.GetExposureRange()
        
    def GetIntTime(self, exp=-1):
        """Gets integration time in microseconds for a given exposure.
    Parameters:
        exp -   Exposure. If not provided, value for current exposure is returned.
    Returns:
        Integration time in microseconds
"""
        return self.interface.GetIntTime(exp)
    
    def reconfigure(self):
        error = None
        if self._want_setexp is not None:
            error = self._setexp(self._want_setexp)                
            self._want_setexp = None
        return error
        

    def SetThreshold(self, val):
        """Sets threshold value for separating performing from non-performing pixels. Roughly corresponds to  signal-to noise ratio under normalization conditions.
    Parameters:
        val -   new value for threshold (default: 10.0, may be changed in defaults file)
    Returns:
        None
    Higher values results in more missing pixels, lower values lead to higher noise.
"""
        self.threshold=float(val)
        self.prepareNorm()
        self.prepareStitch()

    def GetThreshold(self):
        """Gets current threshold value.
    Parameters:
        None
    Returns:
        current threshold value
"""
        return self.threshold

    def SetBG(self,data):
        """Sets data as current backgound data to be used in processing.
    Parameters:
        data    -   numpy.ndaray with data to be used as a background data.
    Returns:
        None
"""
        if data is None:
            self.bg = self.bgNone
            return
        if ( len(data) != self.X_SIZE or len(data[0]) != self.Y_SIZE ) :
            self.bg = self.bgNone
            raise RuntimeError("SetBG::Range mismatch")
        try:
            self.bg = data.copy()
        except:
            self.bg = np.array(data,dtype=np.float32)

    def SelectBG(self, val):
        """Selects a background data from existing list. The data are set as current backgound data to be used in processing.
    Parameters:
        val    -   index of the data in the list (corresponds to exposure)
    Returns:
        None
"""
        try:
            self.SetBG(self.bgList[val].data)
            self.currentBG = val
        except IndexError:
            self.SetBG(None)
            self.currentBG = None
        except TypeError:
            self.SetBG(None)
            self.currentBG = None

    def GetSelectedBG(self):
        """Gets index of a currently selected background data within the list.
    Parameters:
        None
    Returns:
        index of a currently selected background.
"""
        return self.currentBG

    def SetNorm(self,data=None, mask=None):
        """Sets current normalization data to be used in processing.
    Parameters:
        data    -   numpy.ndaray with data to be used for normalization (default:None).
        mask    -   numpy.ndaray with corresponding mask data to be used for normalization (default:None).
    Returns:
        None
"""
        if data is None:
            self.norm = self.normNone
            self.normSrc = self.normNone
            self.mask = self.maskNone
            self.prepareStitch()
            return
        if ( len(data) != self.X_SIZE or len(data[0]) != self.Y_SIZE ) :
            self.norm = self.normNone
            self.normSrc = self.normNone
            self.mask = self.maskNone
            self.prepareStitch()
            raise RuntimeError("SetNorm::Range mismatch")
            return
        else:
            self.normSrc = data
        
        if mask is not None:
            try:
                self.mask = mask.copy()
            except:
                self.mask = np.array(mask,dtype=np.float32)            
        else:
            self.mask = self.maskNone

        self.prepareNorm()
        self.prepareStitch()

    def GetNormList(self):
        """Gets a list of normalizations as dictionary. It includes all possible keys and boolean values indicate whether the corresponding normalizaion is available at the moment.
    Parameters:
        None
    Returns:
        dictionary {"default":True|False, "recorded":True|False}
"""
        return dict(((key, key in self.normDict) for key in DEFAULT_NORM_KEYS))

    def GetSelectedNorm(self):
        """Gets key for currently selected normalization.
    Parameters:
        None
    Returns:
        key for currently selected normalization ("default" or "recorded") or None if none is selected.
    """
        return self.currentNorm

    def SelectNorm(self,val = None):
        """Selects normalization from the dictionary by the key (either "default", "recorded", or None). 
    Parameters:
        val -   key value. If the key is None or does not exists empty normalization is switched off.
    Returns:
        The key value on success or None otherwise.
    """
        if val in self.normDict:
            self.SetNorm(self.normDict[val].data,self.normDict[val].mask)
            self.currentNorm = val
        else:
            self.SetNorm()
            self.currentNorm = None
        return self.currentNorm
        

    def SetAccuLength(self, val):        
        """Sets accumulation length (window size for time-domain filtering). 
    Parameters:
        val -   accumulation length (coerced to [1,100]).
    Returns:
        None

    This method does not turn accumulation on! Use SetAccumulation()
"""
        self.accuLength=int(val)
        if (self.accuLength < 1):
            self.accuLength = 1
        elif (self.accuLength > 100):
            self.accuLength = 100

    def GetAccuLength(self):
        """Gets current accumulation length (window size for time-domain filtering).
    Parameters:
        None
    Returns:
        accumulation length

    The return value does not depend on whether the accumulation is on or off.        
"""
        return self.accuLength
    
    def ResetAccumulation(self):
        """Resets accumulated data.
    Parameters:
        None
    Returns:
        None
"""
        self._reset = True

    def SetAccumulation(self, on=True):
        """Turns accumulation on or off.
    Parameters:
        on  -   True to switch on or False to switch off.
    Returns:
        None
"""
        if on:
            self.processFlags |= ACCUMULATE
        else:
            self.processFlags &= ~ACCUMULATE
        self.ResetAccumulation()

    def SetDifference(self, on = True):
        """Turns on difference mode on or off."""
        if on:
            self.processFlags |= DIFFERENCE
##            self.parity = True
##            self.odd_frame = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=np.float32)
            self.ResetAccumulation()
        else:
            self.processFlags &= ~DIFFERENCE
        
    def LoadConfig(self,zipName=None,pulseCallback=None):
        """Loads background and normalization data from a configuration file and puts in corresponding list and dictionary.
    Parameters:
        zipName  -   filename for a configuration file. If it is not provided, default configuration file is loaded.
        pulseCallback   -       callback to indicate progress (it is called without parameters).
    Returns:
        None

    If filename is provided as a relative path, it is relative to the module folder.
"""

        bgList=[]
        normDict={}

        configZipName = zipName or "data/"+self.interface.get_id()+".zip"

        with zipfile.ZipFile(configZipName) as configZip:
            if callable(pulseCallback):
                pulseCallback()
            for name in DEFAULT_BG_FILENAMES:                
                with configZip.open(name,"r") as f:
                    file_in=io.BytesIO(b"\n".join(f.read().splitlines()))
                    data = ref.RefData(self.X_SIZE,self.Y_SIZE, style = ref.BG)
                    data.load(file_in)
                    file_in.close()
                bgList.append(data)
                if callable(pulseCallback):
                    pulseCallback()

            for label,name in DEFAULT_NORM_FILENAMES.items():
                if name is None:
                    continue
                with configZip.open(name,"r") as f:
                    file_in=io.BytesIO(b"\n".join(f.read().splitlines()))
                    data = ref.RefData(self.X_SIZE,self.Y_SIZE, style = ref.NORM)
                    data.load(file_in)
                    file_in.close()
                normDict[label] = data                
                if callable(pulseCallback):
                    pulseCallback()
                        
        self.bgList = bgList
        self.normDict = normDict
        self.SelectBG(self.GetExposure())
        s = self.GetSelectedNorm()
        if s in self.normDict:
            self.SelectNorm(s)
        else:
            self.SelectNorm('default')
        if callable(pulseCallback):
            pulseCallback()
            

    def SaveConfig(self,filename,pulseCallback=None):
        """Saves complete list of backgrounds and normalization data dictionary to a configuration file.
    Parameters:
        filename  -   filename for a configuration file (required parameter).
        pulseCallback   -       callback to indicate progress (it is called without parameters).
    Returns:
        None

    If filename is provided as a relative path, it is relative to the module folder.
"""
        if callable(pulseCallback):
            pulseCallback()
        with zipfile.ZipFile(filename,"w") as configZip:
            for i,name in enumerate(DEFAULT_BG_FILENAMES):
                
                if i >= len(self.bgList):
                    break
                name = DEFAULT_BG_FILENAMES[i]
                file_out = tempfile.NamedTemporaryFile(delete=False)
                tempfilename = file_out.name
                self.bgList[i].save(file_out)
                file_out.close()
                configZip.write(tempfilename,name,zipfile.ZIP_DEFLATED)
                os.remove(tempfilename)
                if callable(pulseCallback):
                    pulseCallback()

            for  label,name in DEFAULT_NORM_FILENAMES.items():
                if label in self.normDict and name is not None:
                    file_out = tempfile.NamedTemporaryFile(delete=False)
                    tempfilename = file_out.name
                    self.normDict[label].save(file_out)
                    file_out.close()
                    configZip.write(tempfilename,name,zipfile.ZIP_DEFLATED)
                    os.remove(tempfilename)
                    if callable(pulseCallback):
                        pulseCallback()

    def RecordCurrentBG(self,count=30, ticker=None):
        """Record background data for current exposure.
        Parameters:
        count  -   number of repetitions used for averaging (default: 30).
        Returns:
        terasense.ref.RefData object with the background information."""

        data = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=np.float32)
        dur = self.GetIntTime()
        self.processFlags |= NOPROCESS


        if self.threaded:
            self.start()

            cancelled = False
        
        for i in range(count):

            tmp = self.read() 

            np.add(data,tmp,data)
            
            if callable(ticker):
                cont, skip = ticker(dur)
                if not cont:
                    cancelled = True


        if self.threaded:
            self.stop()
        self.processFlags &= ~NOPROCESS

        if cancelled:
            return None
        np.divide(data,count,data)
        
        return ref.RefData(self.X_SIZE,self.Y_SIZE,data=data, style = ref.BG, exposure = self.GetExposure())
   
    def RecordBG(self,count=0,callback=None):
        """Record background data for all available exposures.
    Parameters:
        count  -   number of repetitions used for averaging. If 0 or not provided, then default value is used (30, may be changed in defaults file).
        callback    -   callback to indicate progress. It is called with completed percentage and it is expected to return tuple (continue, skip), where continue is True unless the process should be aborted (see wx.ProgressDialog)
    Returns:
        True if completed successfully, False if canceled
"""
        flag = False
        cancelled = False
        if self.threaded:
            if self.proc_thread is not None and self.proc_thread.is_alive():
                flag = True
                self.stop()
                

        current_exp = self.GetExposure()
        min_exp, max_exp = self.interface.GetExposureRange()
        if not count:
            try:
                count = self.recordingCount
            except AttributeError:
                count = 30
        exposures = list(range(min_exp, max_exp+1))
        total = sum(map(self.GetIntTime,exposures))*count
        pr = progress(total,callback)

        result = []

        for exp in exposures:

            self.SetExposure(exp)

            x = self.RecordCurrentBG(count, pr.increase)
            if x is not None:
                result.append(x)
            else:
                cancelled=True
                break
            
        if not cancelled:   
            self.bgList = result
        self.SetExposure(current_exp)
        
        if flag:
            self.start(resume=True)
        return (not cancelled)



    
    def RecordNorm(self,count=0,callback=None, force=False):
        """Record normalization data.
    Parameters:
        count  -   number of repetitions used for averaging. If 0 or not provided, then default value is used (30, may be changed in defaults file).
        callback    -   callback to indicate progress. It is called with completed percentage and it is expected to return tuple (continur, skip), where continue is True unless the process should be aborted (see wx.ProgressDialog)
    Returns:
        True if completed successfully, False if canceled.
"""
        cancelled = flag = False
        if self.threaded:
            if self.proc_thread is not None and self.proc_thread.is_alive():
                flag = True
                self.stop()
        if not count:
            try:
                count = self.recordingCount
            except AttributeError:
                count = 30

        
        data = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=np.float64)
        squared = np.zeros((self.X_SIZE,self.Y_SIZE),dtype=np.float64)
        pr = progress(count,callback)

## auto-select exposure
        
        exp = current_exp = self.GetExposure()
        inttime = current_inttime = self.GetIntTime()
        min_exp, max_exp = self.interface.GetExposureRange()


        if self.ROI is None:
            maximum = np.max(self.readNormData())
        else:
            maximum = np.max(self.readNormData()[self.ROI[0][0]:self.ROI[1][0]+1,self.ROI[0][1]:self.ROI[1][1]+1])
        if maximum < 0.9*RAW_MAX :
            while exp < max_exp and old_div(self.GetIntTime(exp+1),current_inttime)*maximum  < 0.9*RAW_MAX:
                exp += 1
                if exp == max_exp:
                    break
            self.SetExposure(exp)
        else:
            while maximum  > 0.9*RAW_MAX and exp > min_exp:
                exp -= 1
                self.SetExposure(exp)
                if self.ROI is None:
                    maximum = np.max(self.readNormData())
                else:
                    maximum = np.max(self.readNormData()[self.ROI[0][0]:self.ROI[1][0]+1,self.ROI[0][1]:self.ROI[1][1]+1])
        if self.ROI is None:
            tmp = self.readNormData()
        else:
            tmp = self.readNormData()[self.ROI[0][0]:self.ROI[1][0]+1,self.ROI[0][1]:self.ROI[1][1]+1]
        maximum = np.max(tmp)
        dark = len(np.where(np.absolute(tmp)<0.02)[0])
        inttime = self.GetIntTime()

        if not force:
            if maximum < 0.1*RAW_MAX:
                self.SetExposure(current_exp)
                raise Exception('Not enough illumination power to record proper normalization. May be the radiation source is off?')
            elif maximum > 0.9*RAW_MAX:
                self.SetExposure(current_exp)
                raise Exception('Too strong illumination to record proper normalization.')
            else:
                if ( self.ROI is None):
                    if dark > 0.2*self.X_SIZE*self.Y_SIZE:
                        self.SetExposure(current_exp)
                        raise Exception('Variation of the illumination over the surface is too strong to record proper normalization.')
                else:
                    if dark > 0.2*(self.ROI[1][0]+1-self.ROI[0][0])*(self.ROI[1][1]+1-self.ROI[0][1]) :
                        self.SetExposure(current_exp)
                        raise Exception('Variation of the illumination over the surface is too strong to record proper normalization.')
    
        self.processFlags |= NOPROCESS
        if self.threaded:
            self.start()
##            tmp = self.read()
##            tmp = self.read()
        for i in range(count):
            tmp = self.read()
            if not (self.processFlags & DIFFERENCE) :
                tmp = self.SubstractBG(tmp)
            np.add(data,tmp,data)
            np.square(tmp,tmp)
            np.add(squared,tmp,squared)
            
            cont, skip = pr.increase(1)

            if not cont:
                cancelled = True
                break
        if self.threaded:
            self.stop()
        self.processFlags &= ~NOPROCESS
        
        if not cancelled:   
            np.divide(data,count,data)
            sd2 = np.subtract(np.divide(squared,count),np.square(data))
        
##            strict = (sd2 > 0)
##            np.around(sd2,6,sd2)
            if np.any(sd2 < 0):
                data = np.where(sd2 < 0, 0, data)
                sd2 = np.where(sd2 < 0, 0, sd2)
##                a = np.extract(sd2<0,sd2)
##                raise(RuntimeError("Negative sd**2 recording normalization values?!"))
            test = (sd2 == 0)
##            sd = np.divide(np.sqrt(sd2),np.sqrt(count)
            sd = np.sqrt(sd2)
##            sd = np.where(test,np.divide(np.sqrt(sd2),count),1)
            mask = sd
            mask = np.where(test,0, np.divide(np.absolute(data),sd))
            mask = np.where(np.logical_and(test, data!=0), 99999,   mask)
            if current_inttime < inttime:
                np.multiply(data,old_div(current_inttime,inttime),data)            
            x = ref.RefData(self.X_SIZE,self.Y_SIZE,data=data.astype(np.float32), mask=mask.astype(np.float32), style = ref.NORM, exposure = self.GetExposure())


            self.normDict['recorded'] = x
            

            self.SelectNorm('recorded')

        self.SetExposure(current_exp)
                  
        if flag:
            self.start(resume=True)
        return (not cancelled)


    def MakeNullBGList(self):
        
        min_exp, max_exp = self.interface.GetExposureRange()
        exposures = list(range(min_exp, max_exp+1))
        self.bgList = []

        for exp in exposures:
            self.bgList.append(ref.RefData(self.X_SIZE,self.Y_SIZE,data=self.bgNone, style = ref.BG, exposure = exp))
            


    def MakeNullNormList(self):
        self.normDict ={}
        for  label,name in DEFAULT_NORM_FILENAMES.items():
            self.normDict[label] = ref.RefData(self.X_SIZE,self.Y_SIZE,data=self.normNone.astype(np.float32), mask=self.maskNone.astype(np.float32), style = ref.NORM, exposure = self.GetExposure())
    

    
