from __future__ import division
from __future__ import print_function
from builtins import str
from builtins import range
from builtins import object
from past.utils import old_div
__svn_revision__ = '$Revision: 433 $'

#
# Low level API for multichannel
# T-BUS controller based on Opal Kelly XEM6001 Spartan-6 module
# Wiring:
#                JP3
# ----------------------
# leds         GND  GND  |
#              SRQ  CLK  |
#              MDO       |
#              MDI0 MDI1 |
#              3.3V 3.3V |
#              MDI2 MDI3 |
#                 ...    |
#
# See http://www.opalkelly.com/library/XEM6001-UM.pdf
#

import time
import terasense.lib.log as log

# Master clock
master_clk = 24000000
# The command buffer size
max_cmd_length = 512
# The rx buffer size
max_rx_length  = 2048
# The amount of data that may safely read in streaming mode provided that the data_rdy status is set
rx_buff_chunk  = old_div(max_rx_length, 2)

# Host interface endpoints
class EP(object):
	# control trigger
	trig     = 0x40
	# writable registers
	cfg      = 0x00
	freq     = 0x01
	rst      = 0x02
	cmd      = 0x03
	rx       = 0x04
	skip     = 0x05
	freqx    = 0x06
	# readable registers
	version  = 0x20
	status   = 0x21
	# the start of the register bank with raw bus input
	bus_in_  = 0x30
	# command buffer
	cmd_buff = 0x80
	# receiver buffer
	rx_buff  = 0xa0

	channel_bits = 10
	channel_mask = (1 << channel_bits) - 1

	@staticmethod
	def cfg_cfg(channels, use_leds):
		"""The global config contains the number of active channels and use LEDs flag"""
		assert (EP.channel_mask & channels) == channels
		v = channels - 1
		if use_leds: v |= 1 << 15
		return v

	@staticmethod
	def freq_cfg(div, interval):
		"""
		The frequency config word consists of the bus clock divider and byte transmission interval in usec.
		The resulting bus clock frequency will be 24/div MHz. 
		"""
		assert 0 < div and div <= 256
		assert 0 < interval and interval < 256
		return (div - 1) | (interval << 8)

	@staticmethod
	def freqx_cfg(interval):
		"""
		The extended frequency config word consists of the turbo mode byte transmission interval in usec.
		"""
		assert 0 < interval and interval < 256
		return (interval << 8)

	@staticmethod
	def rst_cfg(rst_time, rst_hold):
		"""
		The reset config word consists of the reset time and clock hold time for hard reset, both in usec.
		"""
		assert 0 < rst_time and rst_time < 256
		assert 0 < rst_hold and rst_hold < 256
		return rst_time | (rst_hold << 8)

	@staticmethod
	def cmd_cfg(cmd_length, turbo = False, b16 = False, loop = False):
		"""
		Command config word contains the command length - 1 in low order bits and transmission looping flag used for self testing
		"""
		assert 0 < cmd_length and cmd_length <= max_cmd_length
		v = cmd_length - 1
		if turbo : v |= 1 << 10
		if b16   : v |= 1 << 11
		if loop  : v |= 1 << 15
		return v

	@staticmethod
	def rx_cfg(rx_length, wait = False, loopback = False):
		"""
		Receiver config word contains the response length - 1 in low order bits, wait non-zero response flag,
		streaming flag and loopback flag used for self testing. The streaming flag will be set if response length
		argument is zero. In such case the receiving must be stopped explicitly by sending TRIG.stop
		"""
		assert 0 <= rx_length and rx_length <= max_rx_length
		v = rx_length - 1 if rx_length > 0 else 1 << 14
		if wait     : v |= 1 << 13
		if loopback : v |= 1 << 15
		return v

class TRIG(object):
	# Trigger bits (index)
	reset = 0  # Initiate bus reset cycle
	srq   = 1  # Initiate bus srq   cycle 
	start = 2  # Start command transmission
	stop  = 8  # Stop command transmission is stream mode
	auto  = 12 # Start auto mode

class STATUS(object):
	# Status bits
	ready        = 1 << 0  # LED1  The bus data input inverted, modules drive it low after successful startup
	reset        = 1 << 1  #       Set to 1 if bus sqr line is driven low
	hold         = 1 << 2  #       Set to 1 if bus clock line is driven low indicating hard bus reset
	active       = 1 << 3  # LED2  Set to 1 if the transmission is in progress
	tx_done      = 1 << 4  # LED3  Set to 1 if the command transmission is completed and we are feeding zeroes to the bus
	stopping     = 1 << 6  # LED4  Set to 1 if transmitter is stopping (completing last byte transmission)
	error        = 1 << 7  # LED8  Set to 1 if transmitter error is detected (due to invalid configuration in particular)
	data_rdy     = 1 << 8  #       Set to 1 if the response data is ready to be read from receiver buffer
	completed    = 1 << 9  # LED6  Set to 1 if the all requested response data were received and transmitter is stopped
	pause        = 1 << 10 # LED5  Set to 1 if the transmitter is paused waiting the buffer to be freed
	rd_unaligned = 1 << 11 #       Set to 1 if read address is not aligned to the chunk boundary
	underrun     = 1 << 15 # LED7  Set to 1 if the client read from receiver buffer more than amount of data available (streaming mode only)
	# Bits indicating fatal error
	failure   = error | underrun

def trigger(dev, what):
	"""Activate trigger given its index"""
	res = dev.ActivateTriggerIn(EP.trig, what)
	if res != dev.NoError:
		raise RuntimeError('trigger %d failed' % what)

def cfg_set(dev, ep, val):
	"""Set configuration word given the endpoint index"""
	if not hasattr(dev, 'cfg'):
		dev.cfg = {}
	if dev.cfg.get(ep, None) != val:
		res = dev.SetWireInValue(ep, val)
		if res != dev.NoError:
			raise RuntimeError('failed to set %#x -> ep %#x' % (val, ep))
		dev.cfg[ep] = val
		dev.cfg[None] = True

def cfg_commit(dev):
	"""Commit configuration set by cfg_set"""
	if hasattr(dev, 'cfg') and dev.cfg.get(None, None):
		del dev.cfg[None]
		log.dbg('committing cfg=%s' % dev.cfg)
		dev.UpdateWireIns()

def version(dev, up=True):
	"""Read controller version and the number of channels"""
	if up: dev.UpdateWireOuts()
	v = dev.GetWireOutValue(EP.version)
	ver = v >> EP.channel_bits
	chs = 1 + (v & EP.channel_mask)
	return (ver, chs)

def status(dev, up=True):
	"""Read controller status"""
	if up: dev.UpdateWireOuts()
	return dev.GetWireOutValue(EP.status)

def find_not_ready(dev, up=True):
	"""Returns the list of not ready channels"""
	ver, chs = version(dev, up)
	n, nr = 0, []
	for i in range((chs + 15)//16):
		mask = dev.GetWireOutValue(EP.bus_in_ + i)
		m = 1
		while n < chs and m < 0x10000:
			if m & mask:
				nr.append(n)
			n += 1
			m <<= 1
	return nr

def cmd_put(dev, buff):
	"""Write command string to the command buffer"""
	if getattr(dev, 'last_cmd', None) == buff:
		return
	dev.last_cmd = buff
	sz = len(buff)
	if sz % 2:
		buff += b'\0'
		sz += 1
	assert 0 < sz and sz <= max_cmd_length
	if dev.lib_version >= '4.2':
		buff = bytearray(buff)
	res = dev.WriteToPipeIn(EP.cmd_buff, buff)
	if res != sz:
		raise RuntimeError('failed to put command: %d bytes, %d bytes written' % (sz, res))

def rx_buff_read_all(dev, sz, chs):
	"""Read data string from the receiver buffer for all channels"""
	sz_ = sz
	if sz_ % 2: sz_ += 1
	assert 0 < sz_ and sz_ <= max_rx_length
	total = sz_ * chs
	buff = bytearray(total) if dev.lib_version >= '4.2' else b'\0' * total
	cnt = dev.ReadFromPipeOut(EP.rx_buff, buff)
	if cnt != total:
		raise RuntimeError('failed to read rx buffer: %d bytes requested, %d bytes read' % (total, cnt))
	return [buff[i*sz_:i*sz_+sz] for i in range(chs)]

def rx_buff_read_all_skipz(dev, sz, chs, msec_tout = None):
	"""Read data string from the receiver buffer for all channels skipping leading zeros"""
	if msec_tout:
		deadline = time.time() + msec_tout/1000.0
	sz_ = sz
	if sz_ % 2: sz_ += 1
	assert 0 < sz_ and sz_ <= max_rx_length
	total = sz_ * chs
	res, buff = bytearray(), bytearray(total)
	while True:
		cnt = dev.ReadFromPipeOut(EP.rx_buff, buff)
		if cnt != total:
			raise RuntimeError('failed to read rx buffer: %d bytes requested, %d bytes read' % (total, cnt))
		if res:
			res += buff
			break
		if any(buff):
			skip = next((i for i, c in enumerate(buff) if c))
			res  = buff[skip:]
			if not skip:
				break
			buff = buff[:skip]
			total = skip
		elif msec_tout and time.time() > deadline:
			raise RuntimeError('timeout (%d msec) waiting data (%d bytes), status %#x' % (msec_tout, total, status(dev)))

	assert res[0]
	assert len(res) == sz_ * chs
	return [res[i*sz_:i*sz_+sz] for i in range(chs)]

def rx_buff_read_stream(dev, chs):
	"""Read chunks of data from the receiver buffer"""
	return rx_buff_read_all(dev, rx_buff_chunk, chs)

def wait_status(dev, set_bits, clr_bits = 0, msec_tout = None):
	"""Wait the particular status bits to be set or cleared"""
	if msec_tout:
		deadline = time.time() + msec_tout/1000.0
	while True:
		st = status(dev)
		if (st & set_bits) == set_bits and (st & clr_bits) == 0:
			break
		if (st & STATUS.failure):
			raise RuntimeError('failure status %#x' % st)
		if msec_tout and time.time() > deadline:
			if (set_bits & STATUS.ready) and not (st & STATUS.ready):
				raise RuntimeError('timeout (%d msec) waiting ready status on channel(s): %s' % (
						msec_tout, find_not_ready(dev, False)))
			else:
				raise RuntimeError('timeout (%d msec) waiting status %#x/%#x, current status %#x' % (
						msec_tout, set_bits, clr_bits & 0xffff, st))

def wait_ready(dev, msec_tout = None):
	"""Wait ready status (after reset or srq)"""
	return wait_status(dev, STATUS.ready, ~STATUS.ready, msec_tout)
