from __future__ import division
from __future__ import print_function
from builtins import str
from builtins import range
from past.utils import old_div
__svn_revision__ = '$Revision: 433 $'

# 
# T-BUS controller - Opal Kelly FPGA module based implementation
#

import struct
import random
import string
from . import tbus_if as base
from . import tbus_defs as tbus
import terasense.lib.log as log
from .ok import *
from .tmod_hlp import *
import os

# See http://www.opalkelly.com/library/FrontPanelAPI/
from opal_kelly import *

class tbus_ctl(base.tbus_ctl):

	# Reset times (usec units)
	rst_time      = 100
	clk_hold_time = 100

	def __init__(self, cfg):
		"""The cfg parameter receive the implementation specific parameters namespace"""
		base.tbus_ctl.__init__(self, cfg)
		self.dev = FrontPanel()
		

	def _init_core(self, port = ''):
		self.dev.lib_version = self.cfg.ok_version
		if self.dev.NoError != self.dev.OpenBySerial(port):
			return False
		return True
	def _close(self):
		return self.dev.Close()
        
	def _init_controller(self, force_load = False, port = ''):
		"""Open Opel Kelly device and initialize it if necessary"""
		if not self._init_core(port=port):
			return False
		if hasattr(self.cfg, 'ok_on_load'):
			self.cfg.ok_on_load(self)
		if force_load or not self.dev.IsFrontPanelEnabled():
			log.notice('loading controller firmware')
			if self.dev.NoError != self.dev.LoadDefaultPLLConfiguration():
				raise RuntimeError('failed to init PLL')
			path = os.path.abspath(os.path.join(os.path.dirname(__file__),self.cfg.ok_fw_path))
			if not os.path.exists(path):
				raise RuntimeError('failed to locate firmware %s' % path)
			log.dbg('using firmware %s' % path)
			with open(path,"rb") as f:
                                data = f.read()
			if self.dev.NoError != self.dev.ConfigureFPGAFromMemory(data):
				raise RuntimeError('failed to init FPGA')
		else:
			log.dbg('firmware is already loaded')
		if not self.dev.IsFrontPanelEnabled():
			raise RuntimeError('front panel is not supported by the firmware')
		self.ver = version(self.dev)
		return True

	def _mk_cmd(self, cmd_code, addr = 0, data = None, data_len = None):
		"""Create and return T-BUS command"""
		if data_len is None:
			data_len = len(data) if data is not None else 0
		assert (data_len == 0) == ((cmd_code & tbus.cmd_has_data_) == 0)
		len1 = data_len - 1 if (cmd_code & tbus.cmd_has_data_) else 0
		assert 0 <= len1 and len1 < tbus.max_data_length
		log.dbg('  BUS %-5s @%x %d bytes' % (tbus.cmd_names[cmd_code], addr, data_len))
		hdr = struct.pack(tbus.hdr_fmt, cmd_code, len1, addr,
			0 if self.cfg.skip_cmd_cookies else int(random.getrandbits(8)), 0, 0)
		if data is None:
			return hdr
		else:
			return hdr + data

	@staticmethod
	def _check_resp_hdr(req, resp, chan):
		"""Check response header. Returns the number of responses."""
		req_  = struct.unpack(tbus.hdr_fmt, req[:tbus.hdr_sz])
		resp_ = struct.unpack(tbus.hdr_fmt, resp[:tbus.hdr_sz])
		
		if req_[:4] != resp_[:4]:
			raise RuntimeError('T-BUS %s returned unexpected header in channel %d: sent %s, received %s' % (
					tbus.cmd_names[req_[0]], chan, repr(req_), repr(resp_)
				))
		if resp_[4]:
			raise RuntimeError('T-BUS %s returned status %d in channel %d (%d responses)' % (
					tbus.cmd_names[req_[0]], resp_[4], chan, resp_[5]
				))
		return resp_[5]

	@staticmethod
	def _check_resp_data(req, resp, chan):
		"""Compare data received with data sent."""
		req_data, resp_data = req[tbus.hdr_sz:], resp[tbus.hdr_sz:len(req)]
		if req_data != resp_data:
			info, nerrs = 'no difference', 0
			req_ = struct.unpack(tbus.hdr_fmt, req[:tbus.hdr_sz])
			for i, req_b in enumerate(req_data):
				if resp_data[i] != req_b:
					if not nerrs:
						info = '[%d] sent=0x%02x, recv=0x%02x' % (i, ord(req_b), ord(resp_data[i]))
					nerrs += 1
			info += ' (%d invalid bytes)' % nerrs
			raise RuntimeError('T-BUS %s + %d bytes returned unexpected data in channel %d: %s' % (
					tbus.cmd_names[req_[0]], len(req), chan, info
				))

	@staticmethod
	def _check_resp_idle(resp, req_len, chan):
		"""Check trailing bytes of the response."""
		
		for i in range(req_len, len(resp)):
			if resp[i] != 0:
				resp_ = struct.unpack(tbus.hdr_fmt, resp[:tbus.hdr_sz])
				raise RuntimeError('T-BUS %s returned unexpected idle byte 0x%x in channel %d' % (
						tbus.cmd_names[resp_[0]], ord(resp[i]), chan
					))

	@staticmethod
	def _check_resp(req, resp, chan, check_data = False):
		"""Check command response. Returns the number of responses."""
		# We expect response to have trailing idle bytes
		assert len(resp) > len(req)
		rcnt = tbus_ctl._check_resp_hdr(req, resp, chan)
		if (check_data): tbus_ctl._check_resp_data(req, resp, chan)
		tbus_ctl._check_resp_idle(resp, len(req), chan)
		return rcnt

	def init_core(self, port=''):
		return self._init_core(port=port)

	def close(self):
                return self._close()
            
	def init(self, force_load = False, port=''):
		"""Initialize controller. Returns True on success, False on failure."""
		if not self._init_controller(force_load=force_load, port = port):
			return False
		
		log.notice('Opal Kelly controller v%d %d channels' % self.ver)
		if self.ver[1] < self.cfg.ok_channels:
			raise RuntimeError('controller supports %d channels, %d channels required' % (self.ver[1], self.cfg.ok_channels))

		log.dbg('initializing')
		div = int(old_div(master_clk, self.cfg.tbus_freq))
		if float(master_clk) / div != self.cfg.tbus_freq:
			raise RuntimeError("can't set bus frequency " + str(self.cfg.tbus_freq))

		self.cfg.f_div = div

		poll_cmd = self._mk_cmd(tbus.cmd_poll, addr=tmod.status_addr, data=b'\x00\xff')
		cfg_set(self.dev, EP.cfg,  EP.cfg_cfg (self.cfg.ok_channels, self.cfg.ok_use_leds))
		cfg_set(self.dev, EP.rst,  EP.rst_cfg (self.rst_time, self.clk_hold_time))
		cfg_set(self.dev, EP.freq, EP.freq_cfg(div, self.cfg.tbus_tx_idle))
		cfg_set(self.dev, EP.freqx,EP.freqx_cfg(self.cfg.tbus_turbo_idle))
		cfg_set(self.dev, EP.cmd,  EP.cmd_cfg(len(poll_cmd)))
		cfg_set(self.dev, EP.rx,   EP.rx_cfg(len(poll_cmd) + 2, wait = True))
		cfg_set(self.dev, EP.skip, 0)
		cfg_commit(self.dev)

		self.bus_reset()
		self.set_wait_tout(self.cfg.tbus_cmd_tout)

		log.dbg('probing modules')
		cmd_put(self.dev, poll_cmd)
		trigger(self.dev, TRIG.start)
		wait_status(self.dev, STATUS.ready | STATUS.tx_done | STATUS.data_rdy | STATUS.completed, STATUS.active, self.cfg.tbus_cmd_tout)
		r = rx_buff_read_all(self.dev, len(poll_cmd) + 2, self.cfg.ok_channels)
		assert len(r) == self.cfg.ok_channels
		chain, mod_version = None, None
		for i, d in enumerate(r):
			d = bytes(d)
			rcnt = self._check_resp(poll_cmd, d, i)
			va, vb = struct.unpack('BB', d[tbus.hdr_sz:tbus.hdr_sz+2])
			if va != vb:
				raise RuntimeError('mixed module versions in channel #%d' % i)
			log.dbg('    #%d: %d modules v.%x' % (i, rcnt, va))
			if chain is None:
				chain = rcnt
				mod_version = va
			else:
				if chain != rcnt:
					raise RuntimeError('unequal number of modules in T-BUS channels')
				if mod_version != va:
					raise RuntimeError('mixed module versions in T-BUS channels')

		if not chain:
			raise RuntimeError('no modules detected')

		self.mod_version = mod_version
		# We are not expecting the number of modules to be changed after initialization
		self.modules = chain * self.cfg.ok_channels
		self.chain   = chain
		# Every module introduces fixed propagation delay so commit it once
		cfg_set(self.dev, EP.skip, (4 if self.mod_version >= 0x12 else 2) * chain)
		cfg_commit(self.dev)
		# Fix config
		if self.mod_version < 0x12:
			self.cfg.tbus_turbo = self.cfg.tbus_16bit_mode = False

		return True

	def info(self, detailed=False):
		"""Returns string describing the controller hardware"""
		if not detailed:
			return 'Opal Kelly controller'
		return 'Opal Kelly %d channels controller %s v.%d, %dx%d modules v.%x' % (
				self.ver[1], self.get_id(), self.ver[0], self.cfg.ok_channels, self.chain, self.mod_version
			)

	def get_dev_count(self):
		"""Returns number of available Opal Kelly controllers"""
		return self.dev.GetDeviceCount()

	def get_ctl_serial(self,i=0):
		"""Returns serial string of the Opal Kelly controller hardware (not id of the Terasense device)"""
		return self.dev.GetDeviceListSerial(i)

	    
	def get_id(self):
		"""Returns id string of the controller hardware"""
		return self.dev.GetDeviceID()

	def set_id(self, idstring):
		"""Sets id string of the controller hardware"""
		self.dev.SetDeviceID(idstring)

	def set_wait_tout(self, msec_tout):
		"""Set wait timeout in milliseconds"""
		self.wt_tout = msec_tout

	def bus_reset(self):
		"""Issue hard reset on the bus"""
		log.dbg('  BUS reset')
		trigger(self.dev, TRIG.reset)
		wait_ready(self.dev, self.cfg.tbus_rst_tout)

	def bus_srq(self):
		"""Issue SRQ on the bus"""
		log.dbg('  BUS srq')
		trigger(self.dev, TRIG.srq)
		wait_ready(self.dev, self.cfg.tbus_srq_tout)

	@staticmethod
	def round_size(sz):
		return sz + (sz & 1) + 2  # make it even and pad with zero bytes

	def _send_cmd(self, cmd_code, addr, data, check_data):
		"""Send command and returns the array of responses"""
		wait = (cmd_code==tbus.cmd_wait)
		cmd = self._mk_cmd(cmd_code, addr, data)
		rx_len = self.round_size(len(cmd))
		cfg_set(self.dev, EP.freq, EP.freq_cfg(self.cfg.f_div, (self.cfg.tbus_tx_idle, self.cfg.tbus_wt_idle)[wait]))
		cfg_set(self.dev, EP.rx,   EP.rx_cfg(rx_len, wait = wait))
		cfg_set(self.dev, EP.cmd,  EP.cmd_cfg(len(cmd), turbo = self.cfg.tbus_turbo, b16 = self.cfg.tbus_16bit_mode))
		cfg_commit(self.dev)
		cmd_put(self.dev, cmd)
		trigger(self.dev, TRIG.start)
		wait_status(
				self.dev,
				STATUS.ready | STATUS.tx_done | STATUS.data_rdy | STATUS.completed, STATUS.active,
				(self.cfg.tbus_cmd_tout, self.wt_tout)[wait]
			)
		r = rx_buff_read_all(self.dev, rx_len, self.cfg.ok_channels)
		assert len(r) == self.cfg.ok_channels
		for i, d in enumerate(r):
			rcnt = self._check_resp(cmd, d, i, check_data)
			if rcnt != self.chain:
				raise RuntimeError('unexpected number of responses in channel %d: expected %d, received %d' % (i, self.chain, rcnt))
		return r

	def bus_cmd(self, cmd_code, addr = 0, data = None):
		"""
		Send single response command. Raise exception if the command status is not success
		or the data received differs from the data sent. Returns the number of responses.
		"""
		assert (cmd_code & tbus.cmd_mult_response_) == 0
		assert cmd_code != tbus.cmd_poll
		self._send_cmd(cmd_code, addr, data, data is not None)
		return self.modules

	def bus_read_raw(self, addr, data_len):
		"""
		Send read command. Raise exception if the command status is not success.
		Returns the array of (number of responses, data) tuples.
		"""
		assert data_len > 0
		assert (data_len & 1) == 0
		resp_sz = data_len * self.chain
		cmd = self._mk_cmd(tbus.cmd_read, addr, None, data_len)
		assert len(cmd) == tbus.hdr_sz
		chan_total = tbus.hdr_sz + resp_sz
		r, chan_total_ = [None] * self.cfg.ok_channels, self.round_size(chan_total)
		stream_mode = chan_total_ > max_rx_length

		log.dbg('      %s %d bytes in %s mode' %
			('streaming' if stream_mode else 'reading', data_len, 'turbo' if self.cfg.tbus_turbo else 'plain'))

		cfg_set(self.dev, EP.freq, EP.freq_cfg(self.cfg.f_div, self.cfg.tbus_tx_idle))
		cfg_set(self.dev, EP.rx,   EP.rx_cfg(0 if stream_mode else chan_total_))
		cfg_set(self.dev, EP.cmd,  EP.cmd_cfg(tbus.hdr_sz, turbo = self.cfg.tbus_turbo, b16 = self.cfg.tbus_16bit_mode))
		cfg_commit(self.dev)
		cmd_put(self.dev, cmd)
		trigger(self.dev, TRIG.start)

		while chan_total_ > 0:
			wait_status(self.dev, STATUS.data_rdy, 0, self.cfg.tbus_cmd_tout)
			if stream_mode:
				chunk_data = rx_buff_read_stream(self.dev, self.cfg.ok_channels)
			else:
				chunk_data = rx_buff_read_all(self.dev, chan_total_, self.cfg.ok_channels)
			assert len(chunk_data) == self.cfg.ok_channels
			for i, resp in enumerate(chunk_data):
				# if r[i] is None: r[i] = []
				if r[i] is None: r[i] = bytearray()
				# r[i].append(resp[:chan_total_])
				r[i] += resp[:chan_total_]
				# r[i] = resp[:chan_total_]
				# r.append(resp[:chan_total_])
				# r[i] = resp
			chan_total_ -= len(resp)

		if stream_mode:
			trigger(self.dev, TRIG.stop)
			wait_status(self.dev, STATUS.ready, STATUS.active, self.cfg.tbus_cmd_tout)

		# chan_data = [str.join(chunks, sep='') for chunks in r]
		
		# for i, data in enumerate(chan_data):
		for i, data in enumerate(r):
			rcnt = self._check_resp_hdr(cmd, data, i)
			if rcnt != self.chain:
				raise RuntimeError('unexpected number of responses in channel %d: expected %d, received %d' % (i, self.chain, rcnt))
			self._check_resp_idle(data, chan_total, i)

		# return (self.modules, str.join([data[tbus.hdr_sz:chan_total] for data in chan_data], sep=''))
		# return (self.modules, r)
		result = bytearray()
		for elem in r:
			result += elem[tbus.hdr_sz:chan_total]
		# return (self.modules, [data[tbus.hdr_sz:chan_total] for data in r])
		return (self.modules, result)

	def bus_read_coherent(self, addr, data_len):
		"""
		Read coherent data from the specified address. Raise exception if the command status
		is not success or the data is not coherent. Returns (r_cnt, data) tuple.
		"""
		data, buff = b'\x00\xff' * data_len, None
		r = self._send_cmd(tbus.cmd_poll, addr, data, False)
		for i, d in enumerate(r):
			resp = d[tbus.hdr_sz:tbus.hdr_sz+len(data)]
			r_or, r_and = resp[::2], resp[1::2]
			if buff is None: buff = r_or
			if r_or != r_and or r_or != buff:
				raise RuntimeError('data is not coherent in channel %d' % i)

		return (self.modules, buff)

	def bus_read_auto(self, data_len, cb):
		"""Run srq/read cycle in auto mode calling supplied callback on each iteration"""
		assert self.ver[0] >= 4
		assert data_len > 0
		assert (data_len & 1) == 0
		resp_sz = data_len * self.chain
		chan_total = tbus.hdr_sz + resp_sz
		r, chan_total_ = [None] * self.cfg.ok_channels, self.round_size(chan_total)
		assert chan_total_ <= max_rx_length
		trigger(self.dev, TRIG.auto)
		stop = False
		while not stop:
			all_data = rx_buff_read_all_skipz(self.dev, chan_total_, self.cfg.ok_channels, self.wt_tout)
			assert len(all_data) == self.cfg.ok_channels
			# stop = cb((self.modules, str.join([chan_data[tbus.hdr_sz:chan_total] for chan_data in all_data], sep='')))
			# stop = cb((self.modules, all_data))
			result = bytearray()
			for elem in all_data:
				result += elem[tbus.hdr_sz:chan_total]
			# stop = cb((self.modules, [chan_data[tbus.hdr_sz:chan_total] for chan_data in all_data]))
			stop = cb((self.modules, result))
