from __future__ import division
from builtins import object
from past.utils import old_div
import numpy as np
import  cv2




class Worker(object):
    """Image generation class. Uses single 2D array of numpy.float [0,1] to produce three-channel numpy.uint8 image.  """
# processFlags
##    FALSECOLOR = 0x1
    SMOOTH     =   0x10
    NEGATIVE   =   0x20
    MEDIAN      = 0x40
    MIRROR  =   0x100
##    ROTATE = 0x200
    DRAW_FRAME =   0x400
    DRAW_GRID =   0x800
    WATERFALL =     0x1000
    DRAW_ROI   = 0x8000
    DEFAULT_FLAGS = SMOOTH

    WF_ADD = 0x1
    WF_ADD_AVERAGE = 0x2
    WF_OVERLAY = 0x4
    
    def __init__(self, size=(32,32), flags = None, wfRatio = None):
        """Initialize object.
    Parameters:
        size    -   dimensions of the image width, height (default: (32,32))
        flags   -   processing flags
        wfSize  -   dimensions of the waterfall image mode. If None, square image is produced

    Processing flags:
        FALSECOLOR  -   produce image in false colors (rainbow) instead of b/w (tinted)
        SMOOTH     -    smooth image (space-domain filtering)
        NEGATIVE   -    invert image
        MEDIAN      -   use median filtering instead of gaussian blurring
        MIRROR  -   mirror image
        DEFAULT_FLAGS = FALSECOLOR | SMOOTH
"""

        self.processFlags = flags or self.DEFAULT_FLAGS

        self.wfModesDict = {"Add":self.WF_ADD,"Add average":self.WF_ADD_AVERAGE}
        
        self.wfMode = self.WF_ADD
        self.wfFlags = 0x0


        self.max, self.average, self.bins = 0,0,[]
        
        self.brightness = 0 # -1, +1
        self.contrast = 1  # 0-inf
        self.SetGamma(1)   # 0-2
        
        self.smoothness = 0.5


        self.SetImgDims(size)
        self.ROI = self.SetROI()


        self.wfSize = (int(size[1]*wfRatio),size[1]) if wfRatio is not None else (old_div(size[1],2),size[1])

         
        self.ntc = 0
        self.dtc = 0

        self.lut = self.generateLUT()

        
        

    def SetImgDims(self, size):
        if isinstance(size,tuple):
            self.size = size        
            self.img = np.empty( self.size+(3,),dtype=np.uint8)
            
    def makeImg(self, data):
        """Generates image according to current settings.
    Parameters:
        data -   input data
    Returns:
        image
"""
##        self.source =  data.copy()
##        self.data   =  self.generateImage(data)
        if (self.processFlags & self.WATERFALL):

            if self.wfMode == self.WF_ADD:
                sample = data.copy()
            elif self.wfMode == self.WF_ADD_AVERAGE:
                sample = np.array([data.mean(0)])
            self.source[sample.shape[0]:,:] =  self.source[0:-sample.shape[0],:]
            self.source[0:sample.shape[0],:] = sample       
        else:
            self.source =  data.copy()
        
        self.data   =  self.generateImage(self.source)
        self.img = self.data2RGB(self.data)
        return self.img

    
    def SetWaterfall(self, on = True, wfRatio = None):
        """Turns waterfall mode on or off.
    Parameters:
        on   -  switch to turn the waterfall mode either on or off. Default = True
    Returns:
        (width, height) - shape of the image after the action

"""
        if on:
            self.processFlags |= self.WATERFALL
            if wfRatio is not None:
                self.wfSize = (int(self.size[1]*wfRatio), self.size[1])
            self.img = np.zeros( self.wfSize+(3,),dtype=np.uint8)
            self.source = np.zeros( self.wfSize,dtype=np.float32)
        else:
            self.processFlags &= ~self.WATERFALL
            self.img = np.zeros( self.size+(3,),dtype=np.uint8)
            self.source = np.zeros( self.size,dtype=np.float32)

            self.wfSize = (int(old_div(self.size[1],2)), self.size[1])
            self.wfSize = (int(self.size[1]/2), self.size[1])

        return self.img.shape[0:2]

    def StartWaterfall(self):
        """Turns waterfall mode on.
    Parameters:
        None
    Returns:
        (width, height) - shape of the image after the action
"""
        return self.SetWaterfall(True)

    def StopWaterfall(self):
        """Turns waterfall mode off.
    Parameters:
        None
    Returns:
        (width, height) - shape of the image after the action"""
        return self.SetWaterfall(False)

    def WaterfallMode(self,mode = None):
        """Sets the kind of waterfall mode.
    Parameters:
        mode   -   mode id. If None, does nothing but queries the current mode.
    Returns:
        mode id

"""
        try:
            self.wfMode = self.wfModesDict[mode]
            return mode
        except:
            for key, value in list(self.wfModesDict.items()):

                if value == self.wfMode:
                    return key        

    def GetWFModes(self):

        return list(self.wfModesDict.keys())
    
    def SetBrightness(self,black,white):
        """Sets brightness using black and white points.
    Parameters:
        black   -   black point value in [0, white)
        white   -   white point value in (black, 1]
    Returns:
        None
"""
        self.brightness = float(black)

    def SetContrast(self,black,white):
        """Sets contrast using black and white points.
    Parameters:
        black   -   black point value
        white   -   white point value 
    Returns:
        None
"""
        self.contrast =  1/float(white-black) if white-black>0.01 else 100


    def SetGamma(self,val):
        """Sets gamma value.
    Parameters:
        val   -   gamma value (gamma > 0)        
    Returns:
        None
"""
        self.gamma = val if val > 0 else 0
        self.gamma_lut = np.clip(np.power(old_div(np.arange(256,dtype=float),255),self.gamma)*255,0,255).astype(np.uint8)

    def SetSmoothness(self, val):
        """Sets smoothness value.
    Parameters:
        val   -   smoothness value (0, 100]        
    Returns:
        None
"""
        self.smoothness=float(val)/100

    def GetSmoothness(self):
        """Gets smoothness value.
    Parameters:
        None   
    Returns:
        smoothness value (0, 100]
"""
        return int(self.smoothness*100)

    def generateLUT(self,mode="fc"):
        """ Generates look-up table to convert a value in 0-255 range into color.
        Parameters:
            mode --- either "fc" (false-color), or "bw" (blue-tinted black&white).
        Returns a tuple with three lookup tables for R, G, and B channels."""

        xxx = np.arange(256,dtype=np.uint8)

        temp = np.zeros((256,1,3),dtype=np.uint8)

        if mode == "bw":
            temp[:,0,0]=120
            temp[:,0,1]=xxx
            temp[:,0,2]=255
            cv2.cvtColor(temp,cv2.COLOR_HLS2RGB, temp)

        elif mode == "fc":
##            temp[:,0,0]=((135.*(255-xxx))/255).astype(np.uint8)     #Hue
##            temp[:,0,1]=255                                         #Saturation
##            temp[:,0,2]= np.where(xxx<32 ,8*xxx, 255)               #Value
##
##            cv2.cvtColor(temp,cv2.COLOR_HSV2RGB, temp)

            temp[0:64,0,0]= 0
            temp[0:64,0,1]=0
            temp[0:64,0,2]= (xxx[0:64]*255./64).astype(np.uint8)

            temp[64:96,0,0]= 0
            temp[64:96,0,1]= (xxx[0:32]*255./32).astype(np.uint8)
            temp[64:96,0,2]= 255

            temp[96:128,0,0]= 0
            temp[96:128,0,1]= 255
            temp[96:128,0,2]= (255 - xxx[0:32]*255./32).astype(np.uint8)

            temp[128:192,0,0]= (xxx[0:64]*255./64).astype(np.uint8)
            temp[128:192,0,1]= 255
            temp[128:192,0,2]= 0

            temp[192:256,0,0]= 255
            temp[192:256,0,1]= (255 - xxx[0:64]*255./63).astype(np.uint8)   # To get (255, 0, 0) and i=255
            temp[192:256,0,2]= 0

        else:
            temp[:,0,0]=xxx
            temp[:,0,1]=xxx
            temp[:,0,2]=xxx
                    
        return (temp[:,0,0].copy(),temp[:,0,1].copy(),temp[:,0,2].copy())

    def SetFC(self,mode=True):
        if mode:
            self.lut = self.generateLUT("fc")
        else:
            self.lut = self.generateLUT("bw")
        
    def generateImage(self,data):
        """Applies image-processing operations to the original data. Data processing is performed in-place.
    Parameters:
        data    -   input data
    Returns:
        output data (in most cases it is the same array).
"""
        data = data.copy()
        if (self.processFlags & self.SMOOTH and self.smoothness > 0) :
            if (self.processFlags & self.MEDIAN):
                k = 3 if self.smoothness <=0.5 else 5
                cv2.medianBlur(data,k,data)
##                cv2.medianBlur(data[self.ROI[0][0]:self.ROI[1][0]+1,self.ROI[0][1]:self.ROI[1][1]+1],k,data[self.ROI[0][0]:self.ROI[1][0]+1,self.ROI[0][1]:self.ROI[1][1]+1])
            else:
                cv2.GaussianBlur(data,(3,3),self.smoothness, data, borderType = cv2.BORDER_REPLICATE)                
##                cv2.GaussianBlur(data[self.ROI[0][0]:self.ROI[1][0]+1,self.ROI[0][1]:self.ROI[1][1]+1],(3,3),self.smoothness, data[self.ROI[0][0]:self.ROI[1][0]+1,self.ROI[0][1]:self.ROI[1][1]+1], borderType = cv2.BORDER_REPLICATE)                

        if (self.processFlags & self.MIRROR):
            data=np.fliplr(data)
##        if (self.processFlags & self.ROTATE):
##            data=np.rot90(data).copy()
        x = np.clip((255*self.contrast*(data-self.brightness)),0,255).astype(np.uint8)
        

        if self.gamma != 1:
            cv2.LUT(x,self.gamma_lut,x)
        if (self.processFlags & self.NEGATIVE):
            np.subtract(255, x, x)
           
        
        
        
        return x

    
    def data2RGB(self,data,contours=None):
        """Converts one-channel BW data to three-channel RGB data, output depends on FALSECOLOR flag.
    Parameters:
        data    -   input data
    Returns:
        A link to self.img, which contains resulting image
    
"""

        temp = np.empty(data.shape+(3,),dtype=np.uint8)        
        temp[:,:,0] = cv2.LUT(data,self.lut[0])
        temp[:,:,1] = cv2.LUT(data,self.lut[1])
        temp[:,:,2] = cv2.LUT(data,self.lut[2])
        return temp




    def statistics(self, selection=(None,None)):
        """Calculates simple statistics on data or a region of the data (mean, minimum, and maximum values).
    Parameters:
        data    -   input data
        selection   -   ((x1,y1), (x2,y2)) coordinates of two corners of the region to be used. If omitted, entire array is used/
    Returns:
        A dictionary {"mean":float, "min":float, "max":float}
"""

        
        x = self.source
        if (self.processFlags & self.MIRROR):
            x = np.fliplr(x)
##        if (self.processFlags & self.ROTATE):
##            x = np.rot90(x)
            
        a, b = selection        
        if isinstance(a,tuple) and isinstance(b,tuple):
            x = x[a[0]:b[0]+1,a[1]:b[1]+1]
        mean = np.mean(x)
        amin = round(np.amin(x),3)
        amax = round(np.amax(x),3)
            
        return {"mean":mean, "min":amin, "max":amax}

    def SetROI(self,roi=None):
        """Sets/resets region of interest in the form of ((x0,y0),(x1,y1)). Parameters have to satisfy conditions x0<=x1, y0<=y1. """
        try:
            if roi[0][0]<0:
                roi[0][0] = 0
            elif roi[0][0]>=self.size[0]:
                roi[0][0] = self.size[0]-1            
            if roi[1][0]<0:
                roi[1][0] = 0
            elif roi[1][0]>=self.size[0]:
                roi[1][0] = self.size[0]-1
                
            if roi[0][1]<0:
                roi[0][1] = 0
            elif roi[0][1]>=self.size[1]:
                roi[0][1] = self.size[1]-1            
            if roi[1][1]<0:
                roi[1][1] = 0
            elif roi[1][1]>=self.size[1]:
                roi[1][1] = self.size[1]-1    
            
            self.ROI = ( (min(roi[0][0],roi[1][0]),min(roi[0][1],roi[1][1])), (max(roi[0][0],roi[1][0]),max(roi[0][1],roi[1][1])) )
        except Exception as e:
            self.ROI = ((0,0),(self.size[0]-1,self.size[1]-1))
        return self.ROI
